/***************************************************************************
main.cpp  -  description
   -------------------
begin                : Fri Dec  6 15:19:06 PST 2002
copyright            : (C) 2002 by Oleksiy Kohany
email                : kohany@girinst.org
***************************************************************************/

/***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************/

#ifdef HAVE_CONFIG_H
   #include <config.h>
#endif

#include <unistd.h>
#include <iostream>
#include <fstream>
#include "functions.h"
#include "cblastaln.h"
#include <string.h>
#include <stdlib.h>

//=================================================================
void usage(char* exec)
{
    if(!exec)
    {
        return;
    }
    if(strrchr(exec, '/'))
    {
        exec = strrchr(exec, '/') + 1;
    }
    cerr << "NAME\n\t" << exec;
#ifdef VERSION

    cerr << " V" << VERSION;
#endif

    cerr << " - blast output parser\n";
    cerr << "\nSYNOPSIS\n\t" << exec << " [blastout] [outmap] [-aln outaln]\n";
    cerr << "\nDESCRIPTION\n\tParses blast report, and saves result as map.\n";
    cerr << "\tAlso can save optional alignments\n";
    cerr << "\nOPTIONS\n";
    cerr << "\t-aln filename\t- saves alignments";
    cerr << "\nFLAGS\n";
    cerr << "\t-noself\t- removes self hits, where name, coordinates and direction are the same.\n";
    cerr << "\t-exp\t- use expectation value for scoring instead of alignment score.\n";
#ifdef COPYRIGHT

    cerr << "\nCOPYRIGHT\n\t" << COPYRIGHT << "\n";
#endif

}

//=================================================================
int main(int argc, char** argv)
{
    istream* is = 0;
    ostream* os = 0;
    ostream* saln = 0;
    bool noself = false;
    char **args = argv;
    ARGMAP amap;
    getoptions(argv, amap, 1, "-aln");
    getflags(argv, amap, 4, "-help", "-version", "-noself", "-exp");
    if(amap.find("-help") != amap.end() || amap.find("-version") != amap.end())
    {
        usage(argv[0]);
        exit(0);
    }
    if((args = assigninput(is, args)) == 0)
    {
        usage(argv[0]);
        exit(10);
    }
    if((args = assignoutput(os, args)) == 0)
    {
        exit(20);
    }
    if(amap.find("-noself") != amap.end())
    {
        noself = true;
    }
    if(amap.find("-aln") != amap.end())
    {
        saln = openofs(amap["-aln"].c_str());
    }
    //-----------------------------------------------------------------
    try
    {
        CBlastAln aln;
        if(amap.find("-exp") != amap.end()) {
        	aln.setEvaluateBy(CBlastAln::BY_EXPECTATION);
        }
        while(aln.read(*is) == 1)
        {
            if(noself && aln.isSelfHit())
            {
                continue;
            }
#ifdef DB2QU1
            aln.printMap_db2(*os);
#else

            aln.printMap_db1(*os);
#endif

            if(saln)
            {
                aln.printAln(*saln);
            }
        }
    }
    catch (int xCode)
    {
        cerr << argv[0] << " : Cought exception: " << "CODE " << xCode << "\n";
        cerr.flush();
        exit(xCode);
    }
    catch (invalid_argument exc)
    {
        cerr << argv[0] << " : " << exc.what() << "\n";
        exit(40);
    }
    //-----------------------------------------------------------------
    if(is != &cin) {
       delete is;
    }
    if(os != &cout) {
       delete os;
    }
    if(saln != NULL)
    {
        saln->flush();
        delete saln;
    }
    return 0;
}
