#ifndef _BIOALIGNMENT_H_
#define _BIOALIGNMENT_H_

#include "BioSequence.h"
#include <stdio.h>
#include <iostream>
#include <vector>
#include "ScoreMatrix.h"
using namespace std;

class AlnStat
{
public:
    long iMMCnt;
    long iTRCnt;
    long iTVCnt;
    long iMCnt; //Matches
    long iPCnt; //Positives
    long iGCnt;
    AlnStat()
    {
        reset();
    }
    void reset()
    {
        iMMCnt = iTRCnt = iTVCnt = iMCnt = iGCnt = iPCnt = 0;
    }
    AlnStat& operator = (const AlnStat& stat)
    {
        iMMCnt = stat.iMMCnt;
        iTRCnt = stat.iTRCnt;
        iTVCnt = stat.iTVCnt;
        iMCnt = stat.iMCnt;
        iGCnt = stat.iGCnt;
        return *this;
    }
    void print(ostream& os, unsigned int iLineLen = 70) const
    {
        os << "m= " << iMCnt << "\tmm= " << (iMMCnt-iGCnt) << "\tts= " << iTRCnt << "\tgs= " << iGCnt << "\ts= " << ((iMCnt+iMMCnt==0) ? -1.0 : (double)iMCnt/(iMCnt+iMMCnt)) << "\tts/tv= " << ((iTVCnt==0) ? -iTRCnt : (double)iTRCnt/iTVCnt) << " #\n";
    }
    void print(FILE* os, unsigned int iLineLen = 70) const
    {
        fprintf(os, "m= %ld\tmm= %ld\tts= %ld\tgs= %ld\ts= %.2f\ttv= %ld\tts/tv= %.2f #\n",  iMCnt, iMMCnt-iGCnt, iTRCnt, iGCnt, ((iMCnt+iMMCnt==0) ? -1.0 : (double)iMCnt/(iMCnt+iMMCnt)), iTVCnt, ((iTVCnt==0) ? -iTRCnt : (double)iTRCnt/iTVCnt));
    }
};

ostream& operator << (ostream& os, const AlnStat& stat);

class BioAlignment
{
protected:
    BioSubSequence m_sCon;
    BioSubSequence m_sSeq;
    double Score;
public:
    BioAlignment();
    BioAlignment(const BioSequence &sCon, const BioSequence &sSeq);
    BioAlignment(const BioSubSequence &sCon, const BioSubSequence &sSeq);
    BioAlignment(const BioAlignment& aln);
    virtual  ~BioAlignment();
    BioAlignment& assign(const BioSequence& sCon, const BioSequence& sSeq);
    BioAlignment& assign(const BioSubSequence& sCon, const BioSubSequence& sSeq);
    BioAlignment& assign(const BioAlignment& aln);
    BioSubSequence& getCon();
    BioSubSequence& getSeq();
    const BioSubSequence& getCon() const;
    const BioSubSequence& getSeq() const;
    virtual double calcSim() const;
    static double calcSim(const BioSequence &sCon, const BioSequence &sSeq, vector<string> const* masks = NULL);
    void printHead(ostream& os);
    void printHead(FILE* FOUT);
    virtual void printAlnHead(FILE* FOUT);
    virtual void printAlnHead(ostream& os);
    void printAln(FILE* FOUT, int iLineLen=LINELENGTH, const char* sBlahS="", const char* sBlahQ="") ;
    void printAln(ostream& os, int iLineLen=LINELENGTH, const char* sBlahS="", const char* sBlahQ="") ;
    void print(FILE* FOUT, unsigned int iLineLen = LINELENGTH);
    void print(ostream& os, unsigned int iLineLen = LINELENGTH);
    void clear();
    virtual int read(istream& is);
    virtual int readPaln(istream& is) throw();
    virtual int readAxt(istream& is) throw();
    virtual void setScore( const double& _newVal);
    virtual const double& getScore() const;
    const AlnStat getStat() const;
    const AlnStat getStat(vector<string> *ignoreList) const;
    static AlnStat getStat(const BioSequence& seq1, const BioSequence& seq2, const vector<string> *ignoreList = NULL);
    static bool isTransition(char c1, char c2);
    static bool isAAMatch(char c1, char c2);
    static bool isNAMatch(char c1, char c2);
    static char getAvg(string& s, int mode = 0);
    void ungap(int minGap=21, ostream *os1 = NULL, ostream *os2 = NULL) throw();
    bool isSelfHit();
    /** No descriptions */
    void addMask(string s);
    /** No descriptions */
    void clearMasks();
    /** Write property of char** matrix. */
    static void loadMatrix(const char* s);
    static const ScoreMatrix& getMatrix();
    double calcScore() const;
    double calcScore(size_t start, size_t end) const;
    static double calcScore(const string& s1, const string& s2, size_t start, size_t end, const ScoreMatrix* scoreMatrix = NULL, vector<double>* outscores = NULL, bool cumulative = false, bool ignoreGaps = false);
    BioAlignment& sharp(int edgeWidht);
    BioAlignment& trimLeft(size_t length);
    BioAlignment& trimRight(size_t length);
private: // Private methods
    vector<string> masks;
    static ScoreMatrix matrix;
    static string lastLoadedMatrix;
    static pthread_mutex_t matrixLoadMutex;
    virtual char getLink(char c1, char c2);
};

ostream& operator << (ostream& os, BioAlignment& aln);

#endif //_BIOALIGNMENT_H_
