/***************************************************************************
                          cblastaln.cpp  -  description
                             -------------------
    begin                : Thu Nov 21 2002
    copyright            : (C) 2002 by Oleksiy Kohany
    email                : kohany@girinst.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <string.h>
#include <stdlib.h>
#include "cblastaln.h"
#include "functions.h"

using namespace std;

const char sQUN[] = "Query=";
const char sDBN[] = ">";
const char sQUL[] = "Query:";
const char sDBL[] = "Sbjct:";
const char sSCR[] = "Score =";
const char sEXP[] = "Expect =";
const char sPRB[] = "P =";
const char sIDT[] = "Identities =";
const char sPOS[] = "Positives =";
const char sERR[] = "ERROR:";
const char sEXC[] = "EXIT CODE";
const char sFAT[] = "FATAL:";
const char sDFA[] = "DFA:";
const char sLN1[] = " letters";
const char sLN2[] = "Length =";


char CBlastAln::sDBName[MAXNAMESIZE] = "";
char CBlastAln::sQUName[MAXNAMESIZE] = "";
char CBlastAln::sReadBuf[MAXLINELENGTH] = "";
unsigned int CBlastAln::iDBLen = 0;
unsigned int CBlastAln::iQULen = 0;
//const short CBlastAln::BY_SCORE = 0;
//const short CBlastAln::BY_EXPECTATION = 1;
/** No descriptions */
CBlastAln::CBlastAln() : evaluateBy(BY_SCORE)
{}
/** No descriptions */
CBlastAln::~CBlastAln()
{}
void CBlastAln::setEvaluateBy(short type) {
	switch(type) {
	case BY_EXPECTATION:
		evaluateBy = type;
		break;
	default:
		evaluateBy = BY_SCORE;
	}
}
/** Reads one HSP (high scoring pair) from blast output file. Information about
query name and database sequence name is stored in static member variables. */
int CBlastAln::read(istream& is)
{
    return read(*this, is);
}
/** Reads one HSP (high scoring pair) from blast output file. Information about
query name and database sequence name is stored in static member variables. */
int CBlastAln::read(FILE* is)
{
    return read(*this, is);
}
/** Reads one HSP from blast output, and assigns it to CBlastAln& object */
int CBlastAln::read(CBlastAln& aln, istream& is)
{
    aln.clear();
    char sLine[MAXLINELENGTH] = "";
    int iF, iT;
    int iCF = 0, iCT = 0, iSF = 0, iST = 0;
    int nLCnt = 0;
    if(aln.readHSPHead(is))
    {
        aln.getCon().setName(sQUName);
        aln.getSeq().setName(sDBName);
        aln.setDbLen(iDBLen);
        aln.setQuLen(iQULen);
        //    aln.setScore(aln.m_iScr);
        //Geting to the first 'Query: ' line in alignment.
        while(strstr(sReadBuf, sQUL) == 0 && !(is.getline(sReadBuf, MAXLINELENGTH)).eof())
            ;
        //Reading line by line
        //    for(; !is.eof(); is.getline(sReadBuf, MAXLINELENGTH)) {
        while(!is.eof())
        {
            //Reading from Query line
            if(sscanf(sReadBuf + strlen(sQUL), "%d %s %d", &iF, sLine, &iT) < 3)
            {
                throw(invalid_argument("CBlastAln::read(CBlastAln&, istream&) : incoplete data line.\n"));
            }
            aln.getCon().append(sLine);
            if(nLCnt == 0)
            {
                //Initializing coordinates
                iCF = iF;
                iCT = iT;
            }
            else
            {
                //Updating coordinates
                iCT = iT;
            }
            //Read one line between Query and Sbjct
            is.getline(sReadBuf, MAXLINELENGTH);

            //In case of short report (-noseqs) just read line is subject's line.
            if(strstr(sReadBuf, sDBL) == 0)
            {
                if(is.getline(sReadBuf, MAXLINELENGTH).fail() || strstr(sReadBuf, sDBL) == 0)
                {
                    return 0;
                }
            }
            //Reading from Sbjct line
            if(sscanf(sReadBuf + strlen(sDBL), "%d %s %d", &iF, sLine, &iT) < 3)
            {
                throw(invalid_argument("CBlastAln::read(CBlastAln&, istream&) : incomplete data line.\n"));
            }
            aln.getSeq().append(sLine);
            if(nLCnt == 0)
            {
                //Initializing coordinates
                iSF = iF;
                iST = iT;
            }
            else
            {
                //Updating coordinates
                iST = iT;
            }
            aln.getCon().setFrom(iCF);
            aln.getCon().setTo(iCT);
            aln.getSeq().setFrom(iSF);
            aln.getSeq().setTo(iST);
            //Read empty lines after Sbjct, or untill next header
            while(!(is.getline(sReadBuf, MAXLINELENGTH)).eof())
            {
                char *ptr;
                if((ptr = strstr(sReadBuf, sQUN)) != 0)
                {
                    sscanf(ptr + strlen(sQUN), "%s", sQUName);
                    iQULen = 0;
                    continue;
                }
                if((ptr = strstr(sReadBuf, sDBN)) != 0)
                {
                    sscanf(ptr + strlen(sDBN), "%s", sDBName);
                    continue;
                }
                if((ptr = strstr(sReadBuf, sLN1)) != 0 && iQULen == 0)
                {
                    ptr = sReadBuf;
                    while(*ptr == ' ')
                    {
                        ptr++;
                    }
                    sscanf(cleanComa(ptr), "(%d", &iQULen);
                    continue;
                }
                if((ptr = strstr(sReadBuf, sLN2)) != 0)
                {
                    sscanf(cleanComa(ptr + strlen(sLN2)), "%d", &iDBLen);
                    continue;
                }
                if(strstr(sReadBuf, sQUL) != 0)
                {
                    break;
                }
                if(strstr(sReadBuf, sSCR) != 0)
                {
                    return 1;
                }
            }
            if(is.eof())
            {
                return 1;
            }
            nLCnt++;
        }
    }
    return 0;
}
/** Reads one HSP from blast output, and assigns it to CBlastAln& object */
int CBlastAln::read(CBlastAln& aln, FILE* is)
{
    aln.getCon().clear();
    aln.getSeq().clear();
    char sLine[MAXLINELENGTH] = "";
    int iF, iT;
    int iCF = 0, iCT = 0, iSF = 0, iST = 0;
    int res = 0;
    int nLCnt = 0;
    if(aln.readHSPHead(is))
    {
        aln.getCon().setName(sQUName);
        aln.getSeq().setName(sDBName);
        aln.setDbLen(iDBLen);
        aln.setQuLen(iQULen);
        //    aln.setScore(aln.m_iScr);
        //Geting to the first 'Query: ' line in alignment.
        //    while((res = fgets(sReadBuf, MAXLINELENGTH, is)) != 0 && strstr(sReadBuf, sQUL) == 0);
        while(strstr(sReadBuf, sQUL) == 0 && (res = readline(sReadBuf, MAXLINELENGTH, is)) == 0)
            ;
        //Reading line by line
        while(res == 0)
        {
            //Reading from Query line
            if(sscanf(sReadBuf + strlen(sQUL), "%d %s %d", &iF, sLine, &iT) < 3)
            {
                throw(invalid_argument("CBlastAln::read(CBlastAln&, istream&) : incoplete data line.\n"));
            }
            aln.getCon().append(sLine);
            if(nLCnt == 0)
            {
                //Initializing coordinates
                iCF = iF;
                iCT = iT;
            }
            else
            {
                //Updating coordinates
                iCT = iT;
            }
            //Read one line between Query and Sbjct
            res = readline(sReadBuf, MAXLINELENGTH, is);

            //In case of short report (-noseqs) just read line is subject's line.
            if(strstr(sReadBuf, sDBL) == 0)
            {
                if((res = readline(sReadBuf, MAXLINELENGTH, is)) != 0 || strstr(sReadBuf, sDBL) == 0)
                {
                    return 0;
                }
            }
            //Reading from Sbjct line
            if(sscanf(sReadBuf + strlen(sDBL), "%d %s %d", &iF, sLine, &iT) < 3)
            {
                throw(invalid_argument("CBlastAln::read(CBlastAln&, istream&) : incomplete data line.\n"));
            }
            aln.getSeq().append(sLine);
            if(nLCnt == 0)
            {
                //Initializing coordinates
                iSF = iF;
                iST = iT;
            }
            else
            {
                //Updating coordinates
                iST = iT;
            }
            aln.getCon().setFrom(iCF);
            aln.getCon().setTo(iCT);
            aln.getSeq().setFrom(iSF);
            aln.getSeq().setTo(iST);
            //Read empty lines after Sbjct, or untill next header
            while(!(res = readline(sReadBuf, MAXLINELENGTH, is)))
            {
                char *ptr;
                if((ptr = strstr(sReadBuf, sQUN)) != 0)
                {
                    sscanf(ptr + strlen(sQUN), "%s", sQUName);
                    iQULen = 0;
                    continue;
                }
                if((ptr = strstr(sReadBuf, sDBN)) != 0)
                {
                    sscanf(ptr + strlen(sDBN), "%s", sDBName);
                    continue;
                }
                if((ptr = strstr(sReadBuf, sLN1)) != 0 && iQULen == 0)
                {
                    ptr = sReadBuf;
                    while(*ptr == ' ')
                    {
                        ptr++;
                    }
                    sscanf(cleanComa(ptr), "(%d", &iQULen);
                    continue;
                }
                if((ptr = strstr(sReadBuf, sLN2)) != 0)
                {
                    sscanf(cleanComa(ptr + strlen(sLN2)), "%d", &iDBLen);
                    continue;
                }
                if(strstr(sReadBuf, sQUL) != 0)
                {
                    break;
                }
                if(strstr(sReadBuf, sSCR) != 0)
                {
                    return 1;
                }
            }
            if(res != 0)
            {
                return 1;
            }
            nLCnt++;
        }
    }
    return 0;
}
/** Forwards read pointer to the begining of next HSP */
bool CBlastAln::readHSPHead(istream& is)
{
    char *ptr;
    int iN;
    bool found = false;
    //  reset();
    do
    {
        if(strstr(sReadBuf, sQUL))
        {
            break;
        }
        if((ptr = strstr(sReadBuf, sQUN)) != 0)
        {
            sscanf(sReadBuf + strlen(sQUN), "%s", sQUName);
            iQULen = 0;
            continue;
        }
        if((ptr = strstr(sReadBuf, sLN1)) != 0 && iQULen == 0)
        {
            ptr = sReadBuf;
            for(ptr = sReadBuf; !isdigit(*ptr) && *ptr != '\0'; ptr++);
            sscanf(cleanComa(ptr), "%d", &iQULen);
            continue;
        }
        if((ptr = strstr(sReadBuf, sLN2)) != 0)
        {
            sscanf(cleanComa(ptr + strlen(sLN2)), "%d", &iDBLen);
            continue;
        }
        if((ptr = strstr(sReadBuf, sDBN)) != 0)
        {
            sscanf(sReadBuf + strlen(sDBN), "%s", sDBName);
            continue;
        }
        if((ptr = strstr(sReadBuf, sSCR)) != 0)
        {
            sscanf(ptr + strlen(sSCR), "%d", &m_iScr);
            if(evaluateBy == BY_SCORE) {
            	setScore(m_iScr);
            }
            found = true;
        }
        if((ptr = strstr(sReadBuf, sEXP)) != 0)
        {
            sscanf(ptr + strlen(sEXP), "%lf", &m_dExp);
            if(evaluateBy == BY_EXPECTATION) {
            	setScore(m_dExp);
            }
        }
        if((ptr = strstr(sReadBuf, sPRB)) != 0)
        {
            sscanf(ptr + strlen(sPRB), "%lf", &m_dPrb);
        }
        if((ptr = strstr(sReadBuf, sIDT)) != 0)
        {
            sscanf(ptr + strlen(sIDT), "%d/%ud", &iN, &m_nLen);
            m_dIdt = ((double)iN) / m_nLen;
        }
        if((ptr = strstr(sReadBuf, sPOS)) != 0)
        {
            sscanf(ptr + strlen(sPOS), "%d/%ud", &iN, &m_nLen);
            m_dPos = ((double)iN) / m_nLen;
        }
#ifdef DEBUG
        if((ptr = strstr(sReadBuf, sDFA)) != 0)
        {
            cerr << sQUName << ": " << sReadBuf << "\n";
        }
#endif
        // Non fatal error:
        if((ptr = strstr(sReadBuf, sERR)) != 0)
        {
            cerr << sReadBuf << "\n";
        }
        // Fatal error:
        if((ptr = strstr(sReadBuf, sFAT)) != 0)
        {
            cerr << sReadBuf << "\n";
        }
        // Exit code:
        if((ptr = strstr(sReadBuf, sEXC)) != 0)
        {
            cerr << sReadBuf << "\n";
            //      int iERRCode = 0;
            //      sscanf(ptr + strlen(sERR), "%d", &iERRCode);
            //      throw(iERRCode);
        }
    }
    while(!is.getline(sReadBuf, MAXLINELENGTH).fail());
    return found;
}
/** Forwards read pointer to the begining of next HSP */
bool CBlastAln::readHSPHead(FILE* is)
{
    char *ptr;
    int iN;
    bool found = false;
    //  reset();
    do
    {
        if(strstr(sReadBuf, sQUL))
        {
            break;
        }
        if((ptr = strstr(sReadBuf, sQUN)) != 0)
        {
            sscanf(sReadBuf + strlen(sQUN), "%s", sQUName);
            iQULen = 0;
            continue;
        }
        if((ptr = strstr(sReadBuf, sDBN)) != 0)
        {
            sscanf(sReadBuf + strlen(sDBN), "%s", sDBName);
            continue;
        }
        if((ptr = strstr(sReadBuf, sLN1)) != 0 && iQULen == 0)
        {
            for(ptr = sReadBuf; !isdigit(*ptr) && *ptr != '\0'; ptr++)
                ;
            sscanf(cleanComa(ptr), "%d", &iQULen);
            continue;
        }
        if((ptr = strstr(sReadBuf, sLN2)) != 0)
        {
            sscanf(cleanComa(ptr + strlen(sLN2)), "%d", &iDBLen);
            continue;
        }
        if((ptr = strstr(sReadBuf, sSCR)) != 0)
        {
            sscanf(ptr + strlen(sSCR), "%d", &m_iScr);
            if(evaluateBy == BY_SCORE) {
            	setScore(m_iScr);
            }
            found = true;
        }
        if((ptr = strstr(sReadBuf, sEXP)) != 0)
        {
            sscanf(ptr + strlen(sEXP), "%lf", &m_dExp);
            if(evaluateBy == BY_EXPECTATION) {
            	setScore(m_dExp);
            }
        }
        if((ptr = strstr(sReadBuf, sPRB)) != 0)
        {
            sscanf(ptr + strlen(sPRB), "%lf", &m_dPrb);
        }
        if((ptr = strstr(sReadBuf, sIDT)) != 0)
        {
            sscanf(ptr + strlen(sIDT), "%d/%ud", &iN, &m_nLen);
            m_dIdt = ((double)iN) / m_nLen;
        }
        if((ptr = strstr(sReadBuf, sPOS)) != 0)
        {
            sscanf(ptr + strlen(sPOS), "%d/%ud", &iN, &m_nLen);
            m_dPos = ((double)iN) / m_nLen;
        }
#ifdef DEBUG
        if((ptr = strstr(sReadBuf, sDFA)) != 0)
        {
            cerr << sQUName << ": " << sReadBuf << "\n";
        }
#endif
        // Non fatal error:
        if((ptr = strstr(sReadBuf, sERR)) != 0)
        {
            cerr << sReadBuf << "\n";
        }
        // Fatal error:
        if((ptr = strstr(sReadBuf, sFAT)) != 0)
        {
            cerr << sReadBuf << "\n";
        }
        // Exit code:
        if((ptr = strstr(sReadBuf, sEXC)) != 0)
        {
            cerr << sReadBuf << "\n";
            //      int iERRCode = 0;
            //      sscanf(ptr + strlen(sERR), "%d", &iERRCode);
            //      throw(iERRCode);
        }
    }
    while(readline(sReadBuf, MAXLINELENGTH,is) == 0);
    return found;
}
/** No descriptions */
istream& operator >> (istream& is, CBlastAln& aln)
{
    aln.read(is);
    return is;
}
/** No descriptions */
ostream& CBlastAln::printAln(ostream &os)
{
    //  CHSPInfo::print(os);
    BioAlignment::print(os);
    return os;
}
/** No descriptions */
void CBlastAln::printMap(ostream &os) const
{
    printMap_db1(os);
}
/** No descriptions */
void CBlastAln::printMap_db1(ostream &os) const
{
    double dSim  = calcSim();
    //cerr << m_sSeq.getName() << "\t" << getDbLen() << "\t"<< m_sCon.getName() << "\t" << getQuLen() << "\n";
    print_ft(os, m_sSeq.getName().c_str(), min(m_sSeq.getFrom(), m_sSeq.getTo()), max(m_sSeq.getFrom(), m_sSeq.getTo()),
             m_sCon.getName().c_str(), min(m_sCon.getFrom(), m_sCon.getTo()), max(m_sCon.getFrom(), m_sCon.getTo()),
             (evaluateBy == BY_EXPECTATION) ? m_dExp : m_iScr, dSim, m_dPos + (dSim - m_dIdt),
             ((m_sSeq.getTo() - m_sSeq.getFrom())*(m_sCon.getTo() - m_sCon.getFrom()) >= 0 ? true : false),
             100.0*(abs(m_sSeq.getTo() - m_sSeq.getFrom())+1)/getDbLen(),
             100.0*(abs(m_sCon.getTo() - m_sCon.getFrom())+1)/getQuLen());
}
/** No descriptions */
void CBlastAln::printMap_db2(ostream &os) const
{
    double dSim  = calcSim();
    print_ft(os, m_sCon.getName().c_str(), min(m_sCon.getFrom(), m_sCon.getTo()), max(m_sCon.getFrom(), m_sCon.getTo()),
             m_sSeq.getName().c_str(), min(m_sSeq.getFrom(), m_sSeq.getTo()), max(m_sSeq.getFrom(), m_sSeq.getTo()),
             (evaluateBy == BY_EXPECTATION) ? m_dExp : m_iScr, dSim, m_dPos + (dSim - m_dIdt),
             ((m_sSeq.getTo() - m_sSeq.getFrom())*(m_sCon.getTo() - m_sCon.getFrom()) >= 0 ? true : false),
             100.0*(abs(m_sCon.getTo() - m_sCon.getFrom())+1)/getQuLen(),
             100.0*(abs(m_sSeq.getTo() - m_sSeq.getFrom())+1)/getDbLen());
}
/** No descriptions */
double CBlastAln::calcSim() const
{
    double dSim = BioAlignment::calcSim();
    return dSim > 0 ? dSim : m_dIdt;
}
char* CBlastAln::cleanComa(char *ptr)
{
    char *comaptr = strchr(ptr, ',');
    if(comaptr != NULL)
    {
        for(; *comaptr != '\0'; comaptr++)
        {
            (*comaptr) = *(comaptr+1);
        }
    }
    return ptr;
}
